@@ -58,11 +58,18 @@ impl Reshape {
         self.state = State::load(&mut self.db);
 
         // Make sure no migration is in progress
-        if let state::Status::InProgress { migrations: _ } = &self.state.status {
+        if let state::Status::InProgress { .. } = &self.state.status {
             println!("Migration already in progress, please complete using 'reshape complete'");
             return Ok(());
         }
 
+        if let state::Status::Completing { .. } = &self.state.status {
+            println!(
+                "Migration already in progress and has started completion, please finish using 'reshape complete'"
+            );
+            return Ok(());
+        }
+
         // Determine which migrations need to be applied by comparing the provided migrations
         // with the already applied ones stored in the state. This will throw an error if the
         // two sets of migrations don't agree, for example if a new migration has been added
@@ -100,14 +107,17 @@ impl Reshape {
             .context("failed to set up helpers")?;
 
         let mut new_schema = Schema::new();
-        let mut processed_migrations: &[Migration] = &[];
+        let mut last_migration_index = usize::MAX;
+        let mut last_action_index = usize::MAX;
         let mut result: anyhow::Result<()> = Ok(());
 
         for (migration_index, migration) in remaining_migrations.iter().enumerate() {
             println!("Migrating '{}':", migration.name);
-            processed_migrations = &remaining_migrations[..migration_index + 1];
+            last_migration_index = migration_index;
 
             for (action_index, action) in migration.actions.iter().enumerate() {
+                last_action_index = action_index;
+
                 let description = action.describe();
                 print!("  + {} ", description);
 
@@ -131,7 +141,20 @@ impl Reshape {
         // If a migration failed, we abort all the migrations that were applied
         if let Err(err) = result {
             println!("A migration failed, aborting migrations that have already been applied");
-            abort_migrations(&mut self.db, processed_migrations)?;
+
+            // Set to the Aborting state. This is to ensure that the failed
+            // migration is fully aborted and nothing is left dangling.
+            // If the abort is interrupted for any reason, the user can try again
+            // by running `reshape abort`.
+            self.state.aborting(
+                remaining_migrations.clone(),
+                last_migration_index + 1,
+                last_action_index + 1,
+            );
+
+            // Abort will only
+            self.abort()?;
+
             return Err(err);
         }
 
@@ -175,9 +198,24 @@ impl Reshape {
 
     pub fn complete_migration(&mut self) -> anyhow::Result<()> {
         // Make sure a migration is in progress
-        let remaining_migrations = match &self.state.status {
-            state::Status::InProgress { migrations } => migrations,
-            state::Status::Applying { migrations: _ } => {
+        let (remaining_migrations, starting_migration_index, starting_action_index) = match self.state.status.clone() {
+            state::Status::InProgress { migrations } => {
+                // Move into the Completing state. Once in this state,
+                // the migration can't be aborted and must be completed.
+                self.state.completing(migrations.clone(), 0, 0);
+                self.state.save(&mut self.db).context("failed to save state")?;
+
+                (migrations, 0, 0)
+            },
+            state::Status::Completing {
+                migrations,
+                current_migration_index,
+                current_action_index
+            } => (migrations, current_migration_index, current_action_index),
+            state::Status::Aborting { .. } => {
+                return Err(anyhow!("migration been aborted and can't be completed. Please finish using `reshape abort`."))
+            }
+            state::Status::Applying { .. } => {
                 return Err(anyhow!("a previous migration unexpectedly failed. Please run `reshape migrate` to try applying the migration again."))
             }
             state::Status::Idle => {
@@ -186,17 +224,9 @@ impl Reshape {
             }
         };
 
-        helpers::tear_down_helpers(&mut self.db).context("failed to tear down helpers")?;
-
-        // Run all the completion changes as a transaction to avoid incomplete updates
-        let mut transaction = self
-            .db
-            .transaction()
-            .context("failed to start transaction")?;
-
         // Remove previous migration's schema
         if let Some(current_migration) = &self.state.current_migration {
-            transaction
+            self.db
                 .run(&format!(
                     "DROP SCHEMA IF EXISTS {} CASCADE",
                     schema_name_for_migration(current_migration)
@@ -205,40 +235,99 @@ impl Reshape {
         }
 
         for (migration_index, migration) in remaining_migrations.iter().enumerate() {
+            // Skip all the migrations which have already been completed
+            if migration_index < starting_migration_index {
+                continue;
+            }
+
             println!("Completing '{}':", migration.name);
 
             for (action_index, action) in migration.actions.iter().enumerate() {
+                // Skip all actions which have already been completed
+                if migration_index == starting_migration_index
+                    && action_index < starting_action_index
+                {
+                    continue;
+                }
+
                 let description = action.describe();
                 print!("  + {} ", description);
 
                 let ctx = MigrationContext::new(migration_index, action_index);
-                let result = action
-                    .complete(&ctx, &mut transaction)
-                    .with_context(|| format!("failed to complete migration {}", migration.name))
-                    .with_context(|| format!("failed to complete action: {}", description));
 
-                if result.is_ok() {
-                    println!("{}", "done".green());
-                } else {
-                    println!("{}", "failed".red());
-                    return result;
+                // Update state to indicate that this action has been completed.
+                // We won't save this new state until after the action has completed.
+                self.state.completing(
+                    remaining_migrations.clone(),
+                    migration_index + 1,
+                    action_index + 1,
+                );
+
+                // This did_save check is necessary because of the borrow checker.
+                // The Transaction which might be returned from action.complete
+                // contains a mutable reference to self.db. We need the Transaction
+                // to be dropped before we can save the state using self.db instead,
+                // which we achieve here by limiting the lifetime of the Transaction
+                // with a new block.
+                let did_save = {
+                    let result = action
+                        .complete(&ctx, &mut self.db)
+                        .with_context(|| format!("failed to complete migration {}", migration.name))
+                        .with_context(|| format!("failed to complete action: {}", description));
+
+                    let maybe_transaction = match result {
+                        Ok(maybe_transaction) => {
+                            println!("{}", "done".green());
+                            maybe_transaction
+                        }
+                        Err(e) => {
+                            println!("{}", "failed".red());
+                            return Err(e);
+                        }
+                    };
+
+                    // Update state with which migrations and actions have been completed.
+                    // Each action can create and return a transaction if they need atomicity.
+                    // We use this transaction to update the state to ensure the action only completes.
+                    // once.
+                    // We want to use a single transaction for each action to keep the length of
+                    // the transaction as short as possible. Wherever possible, we don't want to
+                    // use a transaction at all.
+                    if let Some(mut transaction) = maybe_transaction {
+                        self.state
+                            .save(&mut transaction)
+                            .context("failed to save state after completing action")?;
+                        transaction
+                            .commit()
+                            .context("failed to commit transaction")?;
+
+                        true
+                    } else {
+                        false
+                    }
+                };
+
+                // If the action didn't return a transaction we save the state normally instead
+                if !did_save {
+                    self.state
+                        .save(&mut self.db)
+                        .context("failed to save state after completing action")?;
                 }
             }
 
             println!();
         }
 
+        // Remove helpers which are no longer in use
+        helpers::tear_down_helpers(&mut self.db).context("failed to tear down helpers")?;
+
         self.state
             .complete()
             .context("failed to update state as completed")?;
         self.state
-            .save(&mut transaction)
+            .save(&mut self.db)
             .context("failed to save state after setting as completed")?;
 
-        transaction
-            .commit()
-            .context("failed to apply transaction")?;
-
         Ok(())
     }
 
@@ -324,61 +413,103 @@ impl Reshape {
     }
 
     pub fn abort(&mut self) -> anyhow::Result<()> {
-        let remaining_migrations = match &self.state.status {
-            Status::InProgress { migrations } => migrations,
-            Status::Applying { migrations } => migrations,
+        let (remaining_migrations, last_migration_index, last_action_index) = match self
+            .state
+            .status
+            .clone()
+        {
+            Status::InProgress { migrations } | Status::Applying { migrations } => {
+                // Set to the Aborting state. Once this is done, the migration has to
+                // be fully aborted and can't be completed.
+                self.state.aborting(migrations.clone(), 0, 0);
+                self.state.save(&mut self.db)?;
+
+                (migrations, usize::MAX, usize::MAX)
+            }
+            Status::Aborting {
+                migrations,
+                last_migration_index,
+                last_action_index,
+            } => (migrations, last_migration_index, last_action_index),
+            Status::Completing { .. } => {
+                return Err(anyhow!("Migration completion has already been started. Please run `reshape complete` again to finish it."));
+            }
             Status::Idle => {
                 println!("No migration is in progress");
                 return Ok(());
             }
         };
-        let target_migration = remaining_migrations.last().unwrap().name.to_string();
-
-        helpers::tear_down_helpers(&mut self.db).context("failed to tear down helpers")?;
-
-        // Run all the abort changes as a transaction to avoid incomplete changes
-        let mut transaction = self
-            .db
-            .transaction()
-            .context("failed to start transaction")?;
 
         // Remove new migration's schema
+        let target_migration = remaining_migrations.last().unwrap().name.to_string();
         let schema_name = schema_name_for_migration(&target_migration);
-        transaction
+        self.db
             .run(&format!("DROP SCHEMA IF EXISTS {} CASCADE", schema_name,))
             .with_context(|| format!("failed to drop schema {}", schema_name))?;
 
         // Abort all pending migrations
-        abort_migrations(&mut transaction, remaining_migrations)?;
+        self.abort_migrations(
+            &remaining_migrations,
+            last_migration_index,
+            last_action_index,
+        )?;
+
+        helpers::tear_down_helpers(&mut self.db).context("failed to tear down helpers")?;
 
         self.state.status = state::Status::Idle;
         self.state
-            .save(&mut transaction)
+            .save(&mut self.db)
             .context("failed to save state")?;
 
-        transaction
-            .commit()
-            .context("failed to commit transaction")?;
-
         Ok(())
     }
-}
 
-fn abort_migrations(db: &mut dyn Conn, migrations: &[Migration]) -> anyhow::Result<()> {
-    // Abort all migrations in reverse order
-    for (migration_index, migration) in migrations.iter().enumerate().rev() {
-        print!("Aborting '{}' ", migration.name);
+    fn abort_migrations(
+        &mut self,
+        migrations: &[Migration],
+        upper_migration_index: usize,
+        upper_action_index: usize,
+    ) -> anyhow::Result<()> {
+        // Abort all migrations in reverse order
+        for (migration_index, migration) in migrations.iter().enumerate().rev() {
+            // Skip migrations which shouldn't be aborted
+            // The reason can be that they have already been aborted or that
+            // the migration was never applied in the first place.
+            if migration_index >= upper_migration_index {
+                continue;
+            }
 
-        for (action_index, action) in migration.actions.iter().enumerate().rev() {
-            let ctx = MigrationContext::new(migration_index, action_index);
-            action
-                .abort(&ctx, db)
-                .with_context(|| format!("failed to abort migration {}", migration.name))
-                .with_context(|| format!("failed to abort action: {}", action.describe()))?;
+            print!("Aborting '{}' ", migration.name);
+
+            for (action_index, action) in migration.actions.iter().enumerate().rev() {
+                // Skip actions which shouldn't be aborted
+                // The reason can be that they have already been aborted or that
+                // the action was never applied in the first place.
+                if migration_index == upper_migration_index - 1
+                    && action_index >= upper_action_index
+                {
+                    continue;
+                }
+
+                let ctx = MigrationContext::new(migration_index, action_index);
+                action
+                    .abort(&ctx, &mut self.db)
+                    .with_context(|| format!("failed to abort migration {}", migration.name))
+                    .with_context(|| format!("failed to abort action: {}", action.describe()))?;
+
+                // Update state with which migrations and actions have been aborted.
+                // We don't need to run this in a transaction as aborts are idempotent.
+                self.state
+                    .aborting(migrations.to_vec(), migration_index, action_index);
+                self.state
+                    .save(&mut self.db)
+                    .context("failed to save state")?;
+            }
+
+            println!("{}", "done".green());
         }
-        println!("{}", "done".green());
+        Ok(())
     }
-    Ok(())
 }
 
 pub fn latest_schema_from_migrations(migrations: &[Migration]) -> Option<String> {
